﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text.RegularExpressions;
using Stimulsoft.Base.Drawing;
using Stimulsoft.Report;
using Stimulsoft.Report.Components;
using Stimulsoft.Report.Components.TextFormats;

namespace Summary
{
    public class ReportBuilder
    {
        private bool wrap = false;
        public StiReport PrepareReport(bool wordWrap)
        {
            wrap = wordWrap;
            var report = LoadReport();

            var columns = new[]
            {
                new PrintableColumn
                {
                    ColumnType = ColumnType.Text, DataAlignment = HorizontalAlignment.Left, Name = "Text",
                    Precision = 0, ValuePath = "Text", Width = 50
                },
                new PrintableColumn
                {
                    ColumnType = ColumnType.Number, DataAlignment = HorizontalAlignment.Right, Name = "Number",
                    Precision = 2, ValuePath = "Number", Width = 50
                }
            };
            report.Pages.Add(GeneratePage(report, columns, new string[0], 0, true));

            var data = PrepareData(columns, "Report title", "Report subtitle", report.Pages.Count - 1);
            report.RegBusinessObject("Dane", data);

            return report;
        }

        private static string ToRoman(int number)
        {
            if ((number < 0) || (number > 3999)) throw new ArgumentOutOfRangeException("number");
            if (number < 1) return string.Empty;
            if (number >= 1000) return "M" + ToRoman(number - 1000);
            if (number >= 900) return "CM" + ToRoman(number - 900);
            if (number >= 500) return "D" + ToRoman(number - 500);
            if (number >= 400) return "CD" + ToRoman(number - 400);
            if (number >= 100) return "C" + ToRoman(number - 100);
            if (number >= 90) return "XC" + ToRoman(number - 90);
            if (number >= 50) return "L" + ToRoman(number - 50);
            if (number >= 40) return "XL" + ToRoman(number - 40);
            if (number >= 10) return "X" + ToRoman(number - 10);
            if (number >= 9) return "IX" + ToRoman(number - 9);
            if (number >= 5) return "V" + ToRoman(number - 5);
            if (number >= 4) return "IV" + ToRoman(number - 4);
            if (number >= 1) return "I" + ToRoman(number - 1);
            throw new ArgumentOutOfRangeException("number");
        }

        private StiReport LoadReport()
        {
            var report = new StiReport();
            report.Load("SzablonRaportu.mrt");
            return report;
        }

        private readonly Regex invalidCharactersExpRegex = new Regex(@"[.,_\-\[\]\(\)%/\\\s]");
        private StiPage GeneratePage(StiReport report, PrintableColumn[] columns, IEnumerable<string> groupedColumnsDefinition, int counter, bool sumVisible = false)
        {
            int helperNumber = 0;

            int number = report.Pages.Count;

            StiPage page = (StiPage)report.Pages[0].Clone();
            var segmentWidth = page.Width / page.SegmentPerWidth;

            var header = (StiContainer)page.Components["_naglowek"];
            var dataBand = (StiContainer)page.Components["_dane"];
            var summary = (StiContainer)page.Components["_podsumowanie"];
            var pageSummary = (StiContainer)summary.Components["_podsumowanieFolio"];
            var prevSummary = (StiContainer)summary.Components["_przeniesienieFolio"];
            var total = (StiContainer)summary.Components["_razem"];

            header.Width = page.Width;
            dataBand.Width = page.Width;
            summary.Width = page.Width;
            pageSummary.Width = page.Width;
            prevSummary.Width = page.Width;
            total.Width = page.Width;

            double topFolio = pageSummary.Height;

            #region GroupHeader & GroupFooter
            var groupHeaderBand = (StiContainer)page.Components["_groupHeaderBand"];
            int groupHeaderIndex = page.Components.IndexOf(groupHeaderBand);

            var columnsList = columns.Select(x => new { x.ValuePath, x.Name }).ToList();

            int groupDepth = 1;

            var columnsDefinition = groupedColumnsDefinition as IList<string> ?? groupedColumnsDefinition.ToList();
            foreach (string s in columnsDefinition)
            {
                var groupHeader = (StiGroupHeaderBand)groupHeaderBand.Clone();
                groupHeader.SortDirection = StiGroupSortDirection.None;
                groupHeader.Name = String.Format("groupHeaderBand{0}", s);

                var column = columnsList.First(x => x.ValuePath == s);
                int index = columnsList.IndexOf(column);
                string fieldNameExpression = String.Format("{{Dane.Data.Pole{0}Group}}", index);
                string fieldNameExpression2 = String.Format("{{Dane.Data.Pole{0}GroupText}}", index);

                groupHeader.Condition = new StiGroupConditionExpression(fieldNameExpression);
                var groupHeaderText = (StiText)groupHeader.Components["_groupHeaderText"];
                groupHeaderText.Name = String.Format("_groupHeaderText{0}", s);

                groupHeaderText.Text = columnsDefinition.Count > 1 ? String.Format("{0}. {1}: {2}", ToRoman(groupDepth), column.Name, fieldNameExpression2) : String.Format("{0}: {1}", column.Name, fieldNameExpression2);
                groupHeaderText.VertAlignment = StiVertAlignment.Center;
                groupHeader.Enabled = true;

                var groupHeaderPanel = groupHeader.Components.OfType<StiPanel>().First();
                groupHeaderPanel.Name = String.Format("{0}Panel", groupHeader.Name);

                page.Components.Insert(groupHeaderIndex, groupHeader);
                groupHeaderIndex++;
                groupDepth++;
            }

            var groupFooterBand = (StiContainer)page.Components["_groupFooterBand"];
            int groupFooterIndex = page.Components.IndexOf(groupFooterBand) + 1;

            foreach (string s in columnsDefinition)
            {
                var footerBand = (StiContainer)groupFooterBand.Clone();
                footerBand.Name = String.Format("_groupFooterBand{0}", s);
                page.Components.Insert(groupFooterIndex, footerBand);

            }

            page.Components.Remove(groupHeaderBand);
            page.Components.Remove(groupFooterBand);
            #endregion

            double left = 0.0;
            foreach (PrintableColumn c in columns.ToList())
            {
                string columnName = invalidCharactersExpRegex.Replace(c.Name, "");

                double ratio = (header.Width - left) / columns.Sum(x => x.Width);
                columns = columns.SkipWhile(col => col.Equals(c)).ToArray();

                var columnWidth = c.Width * ratio;
                var right = left + columnWidth;
                if ((int)(left / segmentWidth) != (int)(right / segmentWidth))
                {
                    columnWidth = (int)(right / segmentWidth) * segmentWidth - left;
                }

                StiText headerText = GenerateStiText(
                     c.Name,
                     String.Format("_header{0}{1}", columnName, helperNumber++),
                     new Font("Arial", 9, FontStyle.Bold),
                     columnWidth,
                     header.Height,
                     left,
                     0,
                     new StiGeneralFormatService(),
                     StiTextHorAlignment.Center,
                     true);
                headerText.Border.Color = Color.Black;
                headerText.Border.Size = 1;
                headerText.Border.Side = StiBorderSides.All;
                headerText.Brush = new StiSolidBrush(Color.FromArgb(216, 216, 216));
                header.Components.Add(headerText);

                string dataTextText = String.Format("{{((List<object>)Dane.Data.SpecyficData.BusinessObjectValue)[{0}]}}", counter);

                var wordWrap = true;
                if (c.ColumnType == ColumnType.Number)
                {
                    wordWrap = wrap;
                    dataTextText = String.Format("{{string.Format(\"{{0:N\" + ((List<int>)Dane.Data.PrecisionData.BusinessObjectValue)[{0}] + \"}}\", ((List<object>)Dane.Data.SpecyficData.BusinessObjectValue)[{0}]) + \" \" + ((List<string>)Dane.Data.SufixData.BusinessObjectValue)[{0}]}}", counter);
                }


                StiTextHorAlignment align = StiTextHorAlignment.Width;
                switch (c.DataAlignment)
                {
                    case HorizontalAlignment.Left:
                        align = StiTextHorAlignment.Left;
                        break;
                    case HorizontalAlignment.Center:
                        align = StiTextHorAlignment.Center;
                        break;
                    case HorizontalAlignment.Right:
                        align = StiTextHorAlignment.Right;
                        break;
                }

                StiText dataText = GenerateStiText(
                     dataTextText,
                     String.Format("_data{0}{1}", columnName, helperNumber++),
                     new Font("Arial", 8, FontStyle.Regular),
                     columnWidth,
                     header.Height,
                     left,
                     0, new StiGeneralFormatService(),
                     align,
                     wordWrap);

                dataText.Border.Color = Color.LightGray;
                dataText.Border.Size = 1;
                dataText.Border.Side = StiBorderSides.All;
                dataBand.Components.Add(dataText);

                string pageSummaryTextText = "---";
                string prevPagesTextText = "---";
                string totalTextText = "---";
                string groupedTotalTextText = "---";
                FontStyle groupFontStyle = FontStyle.Regular;

                if (c.ColumnType == ColumnType.Number)
                {
                    var precision = c.Precision;
                    sumVisible = true;

                    string stiPrecisionFormat = String.Format("{{string.Format(\"{{0:N{0}}}\"", precision);
                    var suffix = String.Format("((List<string>)Dane.SumSufixData.BusinessObjectValue)[{0}]", counter);

                    pageSummaryTextText = String.Format("{0}, {1} + \" \" + {2}}}", stiPrecisionFormat, PrepareSumString(TypPodsumowania.Strona, number, counter), suffix);
                    prevPagesTextText = String.Format("{0}, {1} + \" \" + {2}}}", stiPrecisionFormat, PrepareSumString(TypPodsumowania.Przeniesienie, number, counter), suffix);
                    totalTextText = String.Format("{0}, {1} + \" \" + {2}}}", stiPrecisionFormat, PrepareSumString(TypPodsumowania.Razem, number, counter), suffix);
                    groupedTotalTextText = String.Format("{0}, {1} + \" \" + {2}}}", stiPrecisionFormat, PrepareSumString(TypPodsumowania.GrupaRazem, number, counter), suffix);
                    groupFontStyle = FontStyle.Bold;
                }

                if (sumVisible)
                {
                    pageSummary.Enabled = true;
                    prevSummary.Enabled = true;
                    total.Enabled = true;
                    summary.Enabled = true;
                    foreach (var gf in page.Components.OfType<StiGroupHeaderBand>())
                    {
                        StiPanel panel = gf.Components.OfType<StiPanel>().First();
                        panel.Enabled = true;

                        if (gf.Height <= panel.Height)
                        {
                            gf.Height += panel.Height;
                        }
                    }
                }

                StiText podsumowanieFolioText = GenerateStiText(
                    pageSummaryTextText,
                    String.Format("_podsumowanieFolio{0}{1}", columnName, helperNumber++),
                    new Font("Arial", 8, FontStyle.Regular),
                    columnWidth,
                    topFolio,
                    left,
                    0,
                    new StiGeneralFormatService(),
                    align,
                    wordWrap);
                podsumowanieFolioText.Border.Color = Color.Black;
                podsumowanieFolioText.Border.Size = 1;
                podsumowanieFolioText.Border.Side = StiBorderSides.All;
                pageSummary.Components.Add(podsumowanieFolioText);

                StiText przeniesienieFolioText = GenerateStiText(
                    prevPagesTextText,
                    String.Format("_przeniesienieFolio{0}{1}", columnName, helperNumber++),
                    new Font("Arial", 8, FontStyle.Regular),
                    columnWidth,
                    topFolio,
                    left,
                    0,
                    new StiGeneralFormatService(),
                    align,
                    wordWrap);

                przeniesienieFolioText.Border.Color = Color.Black;
                przeniesienieFolioText.Border.Size = 1;
                przeniesienieFolioText.Border.Side = StiBorderSides.All;
                prevSummary.Components.Add(przeniesienieFolioText);

                StiText razemText = GenerateStiText(
                    totalTextText,
                    String.Format("_razem{0}{1}", columnName, helperNumber++),
                    new Font("Arial", 8, FontStyle.Regular),
                    columnWidth,
                    topFolio,
                    left,
                    0,
                    new StiGeneralFormatService(),
                    align,
                    wordWrap);

                razemText.Border.Color = Color.Black;
                razemText.Border.Size = 1;
                razemText.Border.Side = StiBorderSides.All;
                total.Components.Add(razemText);

                foreach (var gf in page.Components.OfType<StiGroupHeaderBand>())
                {
                    var groupFooterPanel = gf.Components.OfType<StiPanel>().First();

                    StiText groupedRezemText = GenerateStiText(
                    groupedTotalTextText,
                    String.Format("{0}Text{1}", gf.Name, c.ValuePath),
                    new Font("Arial", 8, groupFontStyle),
                    columnWidth,
                    topFolio,
                    left,
                    0,
                    new StiGeneralFormatService(),
                    align,
                    wordWrap);
                    groupedRezemText.Brush = new StiSolidBrush(Color.FromArgb(242, 242, 242));
                    groupedRezemText.Border.Color = Color.Black;
                    groupedRezemText.Border.Size = 1;
                    groupedRezemText.Border.Side = StiBorderSides.All;
                    groupFooterPanel.Components.Add(groupedRezemText);
                }

                left += headerText.Width;
                counter++;
            }

            var stiTextFirst = dataBand.Components.OfType<StiText>().First();
            var stiTextLast = dataBand.Components.OfType<StiText>().Last();

            StiBorderSide blackBorderSide = new StiBorderSide(Color.Black, 1, StiPenStyle.Solid);
            StiBorderSide grayBorderSide = new StiBorderSide(Color.LightGray, 1, StiPenStyle.Solid);

            stiTextFirst.Border = new StiAdvancedBorder(grayBorderSide, grayBorderSide, blackBorderSide, grayBorderSide, false, 4, new StiSolidBrush(Color.Black));
            stiTextLast.Border = new StiAdvancedBorder(grayBorderSide, grayBorderSide, grayBorderSide, blackBorderSide, false, 4, new StiSolidBrush(Color.Black));

            double calculatedWidth = dataBand.Components.OfType<StiText>().Sum(component => component.Width);

            double diff = calculatedWidth - header.Width;

            header.Components.OfType<StiText>().Last().Width -= diff;

            stiTextLast.Width -= diff;

            pageSummary.Components.OfType<StiText>().Last().Width -= diff;
            prevSummary.Components.OfType<StiText>().Last().Width -= diff;
            total.Components.OfType<StiText>().Last().Width -= diff;

            Rename(page, number);
            page.Enabled = true;
            return page;
        }

        private StiText GenerateStiText(string text, string name, Font font, double width, double height, double left, double top, StiFormatService textFormat, StiTextHorAlignment alignment, bool wordWrap = false)
        {
            var dataText = new StiText
            {
                Text = text,
                Name = name,
                VertAlignment = StiVertAlignment.Center,
                HorAlignment = alignment,
                Font = font,
                Height = height,
                Margins = { Left = 2 },
                Width = width,
                Top = top,
                Left = left,
                TextFormat = textFormat,
                CanGrow = true,
                GrowToHeight = true,
                WordWrap = wordWrap
            };

            return dataText;
        }

        private enum TypPodsumowania
        {
            Strona,
            Przeniesienie,
            Razem,
            GrupaRazem
        }
        private string PrepareSumString(TypPodsumowania typPodsumowania, int number, int counter)
        {
            var daneString = String.Format("_dane{0}", number);
            var businessObjectValueString = String.Format("((List<object>)Dane.Data.SpecyficData.BusinessObjectValue)[{0}])", counter);
            switch (typPodsumowania)
            {
                case TypPodsumowania.Strona:
                    {
                        var suma = "cSum";
                        return String.Format("{0}({1},{2})", suma, daneString, businessObjectValueString);
                    }
                case TypPodsumowania.Przeniesienie:
                    {
                        var suma1 = "cSumRunning";
                        var suma2 = "cSum";
                        return String.Format("{0}({1},{2} - {3}({1},{2})", suma1, daneString, businessObjectValueString, suma2);
                    }
                case TypPodsumowania.Razem:
                    {
                        var suma = "cSumRunning";
                        return String.Format("{0}({1},{2})", suma, daneString, businessObjectValueString);
                    }
                case TypPodsumowania.GrupaRazem:
                    {
                        var suma = "Sum";
                        return String.Format("{0}({1},{2})", suma, daneString, businessObjectValueString);
                    }
                default:
                    return String.Empty;
            }
        }

        private void Rename(StiComponent component, int number)
        {
            var container = component as StiContainer;
            if (container != null)
            {
                foreach (StiComponent c in container.Components)
                {
                    Rename(c, number);
                }
            }

            component.Name += number;
        }

        private ReportDataContainer PrepareData(PrintableColumn[] columns, string title, string subTitle, int pagesCount)
        {
            var data = GetData(200);

            var Container = new ReportDataContainer
            {
                Title = title,
                SubTitle = subTitle,
                Data = new List<PrintDataContainer>(),
                PagesCount = pagesCount,
                SumPrecisionData = new List<int>(),
                SumSufixData = new List<string>()
            };

            var printColumnDefsCount = columns.Length;
            var groupedColumnsDefs = 0;

            foreach (var c in columns)
            {
                if (c.ColumnType == ColumnType.Number)
                {
                    Container.SumPrecisionData.Add(c.Precision);
                    Container.SumSufixData.Add("");
                }
                else
                {
                    Container.SumPrecisionData.Add(0);
                    Container.SumSufixData.Add(string.Empty);
                }
            }
            
            foreach (var d in data)
            {
                PrintDataContainer cont = new PrintDataContainer
                {
                    SpecyficData = new List<object>(),
                    SufixData = new List<string>(),
                    PrecisionData = new List<int>()
                };

                for (int i = 0; i < printColumnDefsCount; ++i)
                {
                    var c = columns[i];

                    object obj = null;
                    //string prefix = null; nie definiujemy prefixu - tylko sufix i precyzja (jak na razie)
                    string sufix = string.Empty;
                    int decPrec = 0;
                    try
                    {
                        if (c.ColumnType == ColumnType.Number)
                        {
                                obj = d.Number;
                                sufix = "";
                                decPrec = c.Precision;
                        }
                        else
                            obj = d.Text;
                    }
                    catch (NullReferenceException) { }

                    cont.SpecyficData.Add(obj);
                    cont.PrecisionData.Add(decPrec);
                    cont.SufixData.Add(sufix);
                }

                Container.Data.Add(cont);
            }
            return Container;
        }

        private List<DataModel> GetData(int n)
        {
            var data = new List<DataModel>(n);

            for (int i = 0; i < n; i++)
			{
                data.Add(new DataModel { Number = i + 1, Text = String.Format("Text {0}", i + 1)});
			}

            return data;
        }
    }
}
